#include "limiter.h"
#include <lunar/fx.hpp>
#include <lunar/dsp.h>  

#define LIMITER_FILTER_SIZE		64
#define LIMITER_FILTER_MASK		(LIMITER_FILTER_SIZE - 1)

#define assert(x)
#define powf pow

static float LimiterFilter0[LIMITER_FILTER_SIZE/2] = { 0 };
static float LimiterFilter1[LIMITER_FILTER_SIZE/2] = { 0 };
static float LimiterFilter2[LIMITER_FILTER_SIZE/4] = { 0 };
static float LimiterFilter3[LIMITER_FILTER_SIZE/4] = { 0 };

struct sample
{
	float L;
	float R;
};

class Limiter
{
private:
	void CreateFilter(float *p, int size, int shape)
	{
		float sum = 0;

		for (int i = 0; i < size/2; i++)
		{
			if (shape == 0)
				p[i] = (float)sin((i + 1) * M_PI / (size+1));
			else if (shape == 1)
				p[i] = (float)(i + 1);

			sum += p[i];
		}

		for (int i = 0; i < size/2; i++)
			p[i] /= 2 * sum;
	}

public:
	Limiter()
	{
		if (LimiterFilter0[0] == 0.0f)
		{
			CreateFilter(LimiterFilter0, LIMITER_FILTER_SIZE, 0);
			CreateFilter(LimiterFilter1, LIMITER_FILTER_SIZE, 1);
			CreateFilter(LimiterFilter2, LIMITER_FILTER_SIZE/2, 0);
			CreateFilter(LimiterFilter3, LIMITER_FILTER_SIZE/2, 1);

		}

		SetCeiling(1.0f);
		SetThreshold(0.1f);
		gain = 1.0f;
		holdcount = 0;
		fipos = 0;
		inpos = 0;
		rstep = 0.0f;
		rstepexp = 1.0f;
		rtime = 1000;
		mingain = 1.0f;
		lookahead = 127;
		releaseshape = 0;

		filtertype = -1;
		SetFilterType(0);
	}

	void SetCeiling(float x)
	{
		ceiling = x * 32767.0f;
	}

	void SetThreshold(float x)
	{
		threshold = x * 32767.0f;
	}

	void SetRelease(float x)
	{
		rtime = max(1, (int)x);
	}

	void SetFilterType(int x)
	{
		if (x == filtertype)
			return;

		filtertype = x;

		switch(x)
		{
		case 0: pfilter = LimiterFilter0; break;
		case 1:	pfilter = LimiterFilter1; break;
		case 2:	pfilter = LimiterFilter2; break;
		case 3:	pfilter = LimiterFilter3; break;
		case 4:	pfilter = LimiterFilter0; break;
		case 5:	pfilter = LimiterFilter0; break;
		default: assert(false);
		}

		switch(x)
		{
		case 0:
		case 1:
		case 4:
		case 5:
			filtersize = LIMITER_FILTER_SIZE;
			filtermask = LIMITER_FILTER_SIZE - 1;
			break;
		case 2:
		case 3:
			filtersize = LIMITER_FILTER_SIZE / 2;
			filtermask = LIMITER_FILTER_SIZE / 2 - 1;
			break;
		default: assert(false);
		}

		for (int i = 0; i < LIMITER_FILTER_SIZE; i++)
		{
			fi[i] = 0.0f;
			inputs[i].L = 0.0f;
			inputs[i].R = 0.0f;
		}

	}

	void SetLookahead(int x)
	{
		lookahead = x;
	}

	void SetReleaseShape(int x)
	{
		releaseshape = x;
	}

	float *GetMinGainPtr() { return &mingain; }
	int GetLatency() { return lookahead * (filtersize - 1) / 127; }

	void Process(float* lins, float *rins, float* louts, float* routs, int samplesize, int numsamples)
	{
		float outgain = ceiling / threshold;
		
		int ioffset = (filtersize - 1) - lookahead * (filtersize - 1) / 127;

		for (int i = 0; i < numsamples; i++)
		{
			sample in = { *lins, *rins };
			float sum;

			if (filtertype == 5)
			{
				sum = fi[fipos & filtermask];	// no filter
			}
			else
			{
				sum = 0.0f;
	
				for (int j = 0; j < filtersize/2; j++)
					sum += pfilter[j] * (fi[(fipos-1-j) & filtermask] + fi[(fipos + j) & filtermask]);
			}

			if (filtertype == 4)
			{
				*louts = *routs = sum * 32767.0f;
			}
			else
			{
				*louts = max(min(inputs[(inpos + ioffset) & filtermask].L * sum, threshold), -threshold) * outgain;
				*routs = max(min(inputs[(inpos + ioffset) & filtermask].R * sum, threshold), -threshold) * outgain;
			}

			inputs[inpos] = in;

			inpos++;
			inpos &= filtermask;
			
			lins += samplesize;
			rins += samplesize;
			louts += samplesize;
			routs += samplesize;

			mingain = min(mingain, sum);

			float absin = max(fabs(in.L), fabs(in.R));

			if (absin > threshold)
			{
				float g = threshold / absin;
				if (g < gain)
				{
					gain = g;
					holdcount = filtersize;
				}
			}

			fi[fipos & filtermask] = gain;
			fipos++;


			if (holdcount > 0)
			{
				if (--holdcount == 0)
				{
					rstep = (1.0f - gain) / rtime;
					rstepexp = powf(1.0f / gain, 1.0f / rtime);
				}

			}
			else
			{
				if (gain < 1.0f)
				{
					if (releaseshape == 0)
						gain += rstep;
					else
						gain *= rstepexp;

					if (gain > 1.0f)
						gain = 1.0f;
				}
			}
			

		}
	}

	void Idle()
	{
		mingain = 1.0f;
	}


private:
	float ceiling;
	float threshold;
	float gain;
	int holdcount;
	float rstep;
	float rstepexp;
	int rtime;
	int fipos;
	int inpos;
	int filtersize;
	int filtermask;
	int filtertype;
	int lookahead;
	int releaseshape;
	float *pfilter;
	float mingain;
	float fi[LIMITER_FILTER_SIZE];
	sample inputs[LIMITER_FILTER_SIZE];

};



inline double DBToAmplitude(double const a)
{
	return pow(10, (a * 0.05));
}

inline double ExpIntp(double const x, double const y, double const a)
{
        assert((x > 0 && y > 0) || (x < 0 && y < 0));
        double const lx = log(x);
        double const ly = log(y);
        return exp(lx + a * (ly - lx));
}

class jeskolalimiter : public lunar::fx<jeskolalimiter> {
public:
	Limiter limiter;
	bool IdleMode;
	int IdleCount;

	void init() {
		IdleMode = true;
		IdleCount = 0;
	}

	void exit() {
		delete this;
	}

	void process_events() {
		if (globals->threshold)
			limiter.SetThreshold((float)DBToAmplitude(30.0 * (*globals->threshold / 127.0 - 1.0)));
	
		if (globals->ceiling)
			limiter.SetCeiling((float)DBToAmplitude(30.0 * (*globals->ceiling / 127.0 - 1.0)));
	
		if (globals->release)
			limiter.SetRelease((float)ExpIntp(0.01, 1000.0, *globals->release / 127.0) * transport->samples_per_second / 1000.0f);
	
		if (globals->filtertype)
			limiter.SetFilterType(*globals->filtertype);
	
		if (globals->lookahead)
			limiter.SetLookahead(*globals->lookahead);
	
		if (globals->releaseshape)
			limiter.SetReleaseShape(*globals->releaseshape);
	}

	void process_stereo(float *inL, float *inR, float *outL, float *outR, int n) {
		/*if (mode & WM_READ)
		{
			 IdleMode = false;
			 IdleCount = 0;
		}
		else
		{
			 if (IdleMode)
			 {
				 limiter.Idle();
				 return false;
			 }
			 else
			 {
				 IdleCount += numsamples;
				 if (IdleCount >= 1024)
					 IdleMode = true;
	
				 memset(psamples, 0, 2 * sizeof(float) * numsamples);
			 }
		}*/
	
		limiter.Process(inL, inR, outL, outR, 1, n);
	}
	
	int get_latench() {
		return limiter.GetLatency();
	}

};

lunar_fx *new_fx() { return new jeskolalimiter(); }
/*
char const *mi::DescribeValue(int const param, int const value)
{
	static char txt[16];

	switch(param)
	{
	case 0:        
	case 1:        
		sprintf_s(txt, 16, "%.1fdB", 30.0 * (value / 127.0 - 1));
		break;
	
	case 2:       
		sprintf_s(txt, 16, "%.1fms", ExpIntp(0.01, 1000.0, value / 127.0));
		break;

	case 3:
		switch(value)
		{
		case 0: return "sin64";
		case 1: return "lin64";
		case 2: return "sin32";
		case 3: return "lin32";
		case 4: return "silly";
		case 5: return "off";
		default:
			return NULL;
		};
		break;

	case 4:
		sprintf_s(txt, 16, "%.1f%%", 100.0 * value / 127.0);
		break;

	case 5:
		switch(value)
		{
		case 0: return "lin";
		case 1: return "exp";
		default:
			return NULL;
		}
		break;

		

	default:
		return NULL;
	}

	return txt;
}
*/


